// I/O through <stdio.h> or <iostream.h>

#ifndef _CL_IO_H
#define _CL_IO_H

#include "cl_number.h"

#if !defined(CL_IO_STDIO) && !defined(CL_IO_IOSTREAM)
// Prefer stdio based I/O - hopefully less buggy.
#define CL_IO_STDIO
#endif


#if defined(CL_IO_STDIO)

// I/O through <stdio.h>

#include <stdio.h>

#define cl_istream  FILE *
#define cl_ostream  FILE *
#define cl_stdin   stdin
#define cl_stdout  stdout
#define cl_stderr  stderr

inline void fprintchar (cl_ostream stream, char c)
{
	// fputc(c,stream);
	putc(c,stream);
}

inline void fprint (cl_ostream stream, const char * string)
{
	// fprintf(stream,"%s",string);
	fputs(string,stream);
}

inline void fprintdecimal (cl_ostream stream, int x)
{
	fprintf(stream,"%d",x);
}

#endif


#if defined(CL_IO_IOSTREAM)

// I/O through <iostream.h>

#include <iostream.h>

#define cl_istream  istream&
#define cl_ostream  ostream&
#define cl_stdin   cin
#define cl_stdout  cout
#define cl_stderr  cerr

inline void fprintchar (cl_ostream stream, char c)
{
	stream.put(c);
}

inline void fprint (cl_ostream stream, const char * string)
{
	stream << string;
}

inline void fprintdecimal (cl_ostream stream, int x)
{
	stream << dec << x;
}

#endif


inline void fprint (cl_ostream stream, const cl_number& x)
{
	extern void print_number (cl_ostream, const cl_number&);
	print_number(stream,x);
}


#endif /* _CL_IO_H */
