// cl_I internals

#ifndef _CL_I_H
#define _CL_I_H

#include "cl_number.h"
#include "cl_integer.h"
#include "cl_macros.h"
#include "cl_malloc.h"
#include "cl_DS.h"

// maximal needed length of a digit sequence for a fixnum
  #define FN_maxlength  ceiling(cl_value_len,intDsize)
// maximal needed length (without sign) of a digit sequence for a fixnum
  #define pFN_maxlength  ceiling(cl_value_len-1,intDsize)
// minimum length of a digit sequence for a bignum
  #define bn_minlength  ceiling((cl_value_len+1),intDsize)
  // Because bignums with n < ceiling((cl_value_len+1)/intDsize) digits are
  // integers with at most n*intDsize < cl_value_len+1 bits, so they fit into
  // a fixnum, including the sign bit.
  // 1 <= bn_minlength <= 5.
// We have pFN_maxlength <= FN_maxlength <= bn_minlength.

// Private fixnum constructor.
inline cl_I::cl_I (struct cl_fixnum * null, cl_uint w)
{
	unused null;
	word = w;
}
inline cl_I cl_I_from_word (cl_uint w)
{
	return cl_I((struct cl_fixnum *) 0, w);
}

struct cl_bignum : cl_heap {
	unsigned int length;	// length (in digits)
	uintD data[1];		// number in two's complement representation
};

inline cl_bignum* TheBignum (cl_bignum* p)
	{ return p; }
inline cl_bignum* TheBignum (const cl_number& obj)
	{ return (cl_bignum*)(obj.pointer); }

#undef offsetof
#define offsetof(type,ident)  ((unsigned long)&(((type*)0)->ident))

inline cl_bignum* allocate_bignum (unsigned int length)
{
	cl_bignum* p = (cl_bignum*) cl_malloc_hook(offsetof(cl_bignum,data[length]));
	p->refcount = 1;
	p->typetag = cl_typetag_bignum;
	p->length = length;
	return p;
}

// Private constructor.
// ptr should be the result of some allocate_bignum() call.
inline cl_I::cl_I (cl_bignum* ptr)
	{ pointer = ptr; }

// Both work, but the first definition results in less compiler-generated
// temporaries.
#if 1
  #define Bignum  cl_bignum*
#else
  #define Bignum  cl_I
#endif

// Type tests.
inline cl_boolean integerp (const cl_I& x)
	{ unused x; return cl_true; }
inline cl_boolean fixnump (const cl_I& x)
	{ return (cl_boolean) !x.pointer_p(); }
inline cl_boolean bignump (const cl_I& x)
	{ return x.pointer_p(); }

// Sign test:

// (MINUSP x) == (< x 0)
inline cl_boolean minusp (const cl_I& x)
{
	if (fixnump(x))
		// This assumes cl_value_shift + cl_value_len == cl_word_size.
		return (cl_boolean)((cl_sint) x.word < 0);
	else
		return (cl_boolean)((sintD)TheBignum(x)->data[0] < 0);
}

// (ZEROP x) == (= x 0)
inline cl_boolean zerop (const cl_I& x)
{
	return (cl_boolean)(x.word == cl_combine(cl_FN_tag,0));
}

// (EQ x y) == (= x y), assuming y a fixnum
inline cl_boolean eq (const cl_I& x, sint32 y)
{
	return (cl_boolean)(x.word == cl_combine(cl_FN_tag,y));
}


// Umwandlungsroutinen Integer <--> Longword:

// Wandelt Fixnum in Longword um.
// FN_to_L(obj)
// > obj: ein Fixnum
// < ergebnis: der Wert des Fixnum als 32-Bit-Zahl.
inline sint32 FN_to_L (const cl_I& x)
{
	// This assumes cl_value_shift + cl_value_len == cl_word_size.
	return (cl_sint)(x.word) >> cl_value_shift;
}

// FN_L_zerop(x,x_) stellt fest, ob x = 0 ist.
// Dabei ist x ein Fixnum und x_ = FN_to_L(x).
  #define FN_L_zerop(x,x_)  (x_==0)

// FN_L_minusp(x,x_) stellt fest, ob x < 0 ist.
// Dabei ist x ein Fixnum und x_ = FN_to_L(x).
  #define FN_L_minusp(x,x_)  (x_<0)

// Wandelt Integer >=0 in Unsigned Longword um.
// I_to_UL(obj)
// > obj: ein Objekt, sollte ein Integer >=0, <2^32 sein
// < ergebnis: der Wert des Integer als 32-Bit-Zahl.
  extern uint32 I_to_UL (const cl_I& obj);

// Wandelt Integer in Signed Longword um.
// I_to_L(obj)
// > obj: ein Objekt, sollte ein Integer >=-2^31, <2^31 sein
// < ergebnis: der Wert des Integer als 32-Bit-Zahl.
  extern sint32 I_to_L (const cl_I& obj);

#ifdef intQsize

// Wandelt Integer >=0 in Unsigned Quadword um.
// I_to_UQ(obj)
// > obj: ein Objekt, sollte ein Integer >=0, <2^64 sein
// < ergebnis: der Wert des Integer als 64-Bit-Zahl.
  extern uint64 I_to_UQ (const cl_I& obj);

// Wandelt Integer in Signed Quadword um.
// I_to_Q(obj)
// > obj: ein Objekt, sollte ein Integer >=-2^63, <2^63 sein
// < ergebnis: der Wert des Integer als 64-Bit-Zahl.
  extern sint64 I_to_Q (const cl_I& obj);

#endif

// Wandelt Longword in Fixnum um.
// L_to_FN(wert)
// > wert: Wert des Fixnums, ein signed 32-Bit-Integer
//         >= -2^(cl_value_len-1), < 2^(cl_value_len-1)
// < ergebnis: Fixnum mit diesem Wert.
  inline cl_I L_to_FN (sint32 wert)
  {
	return cl_I_from_word(cl_combine(cl_FN_tag,wert));
  }

// Wandelt Longword in Integer um.
// L_to_I(wert)
// > wert: Wert des Integers, ein signed 32-Bit-Integer.
// < ergebnis: Integer mit diesem Wert.
#if (cl_value_len >= 32)
  #define L_to_I(wert)  L_to_FN(wert)
#else
  extern cl_I L_to_I (sint32 wert);
#endif

// Wandelt Unsigned Longword in Fixnum >=0 um.
// UL_to_FN(wert)
// > wert: Wert des Integers, ein unsigned 32-Bit-Integer < 2^(cl_value_len-1).
// < ergebnis: Fixnum mit diesem Wert.
  inline cl_I UL_to_FN (uint32 wert)
  {
	return cl_I_from_word(cl_combine(cl_FN_tag,wert));
  }

// Wandelt Unsigned Longword in Integer >=0 um.
// UL_to_I(wert)
// > wert: Wert des Integers, ein unsigned 32-Bit-Integer.
// < ergebnis: Integer mit diesem Wert.
#if (cl_value_len > 32)
  #define UL_to_I(wert)  UL_to_FN(wert)
#else
  extern cl_I UL_to_I (uint32 wert);
#endif

// Wandelt Doppel-Longword in Integer um.
// L2_to_I(wert_hi,wert_lo)
// > wert_hi|wert_lo: Wert des Integers, ein signed 64-Bit-Integer.
// < ergebnis: Integer mit diesem Wert.
  extern cl_I L2_to_I (sint32 wert_hi, uint32 wert_lo);

// Wandelt Unsigned Doppel-Longword in Integer um.
// UL2_to_I(wert_hi,wert_lo)
// > wert_hi|wert_lo: Wert des Integers, ein unsigned 64-Bit-Integer.
// < ergebnis: Integer mit diesem Wert.
  extern cl_I UL2_to_I (uint32 wert_hi, uint32 wert_lo);

#ifdef intQsize

// Wandelt Quadword in Integer um.
// Q_to_I(wert)
// > wert: Wert des Integers, ein signed 64-Bit-Integer.
// < ergebnis: Integer mit diesem Wert.
  extern cl_I Q_to_I (sint64 wert);

// Wandelt Unsigned Quadword in Integer >=0 um.
// UQ_to_I(wert)
// > wert: Wert des Integers, ein unsigned 64-Bit-Integer.
// < ergebnis: Integer mit diesem Wert.
  extern cl_I UQ_to_I (uint64 wert);

#endif

// Liefert die Differenz x-y zweier Unsigned Longwords x,y als Integer.
// minus(x,y)
inline cl_I minus (uintL x, uintL y)
{
#if (cl_word_size==64) // defined(__alpha__)
	return Q_to_I((sintQ)(uintQ)x - (sintQ)(uintQ)y);
#else
	return L2_to_I( (x<y ? -1 : 0), x-y );
#endif
}


// Umwandlungsroutinen Digit sequence <--> Longword:

// Holt die nchsten pFN_maxlength Digits in ein uint32.
inline uint32 pFN_maxlength_digits_at (uintD* ptr)
{
#if (pFN_maxlength==1)
	return (uint32)(ptr[0]);
#elif (pFN_maxlength==2)
	return ((uint32)(ptr[0])<<intDsize) | (uint32)(ptr[1]);
#elif (pFN_maxlength==3)
	return ((((uint32)(ptr[0])<<intDsize) | (uint32)(ptr[1]))<<intDsize) | (uint32)(ptr[2]);
#elif (pFN_maxlength==4)
	return ((((((uint32)(_ptr[0])<<intDsize) | (uint32)(_ptr[1]))<<intDsize) | (uint32)(_ptr[2]))<<intDsize) | (uint32)(_ptr[3]);
#endif
}

// Schreibt ein uint32 in die nchsten pFN_maxlength Digits.
inline void set_pFN_maxlength_digits_at (uintD* ptr, uint32 wert)
{
#if (pFN_maxlength==1)
	ptr[0] = (uintD)wert;
#elif (pFN_maxlength==2)
	ptr[0] = (uintD)(wert>>intDsize);
	ptr[1] = (uintD)(wert);
#elif (pFN_maxlength==3)
	ptr[0] = (uintD)(wert>>(2*intDsize));
	ptr[1] = (uintD)(wert>>intDsize);
	ptr[2] = (uintD)(wert);
#elif (pFN_maxlength==4)
	ptr[0] = (uintD)(wert>>(3*intDsize));
	ptr[1] = (uintD)(wert>>(2*intDsize));
	ptr[2] = (uintD)(wert>>intDsize);
	ptr[3] = (uintD)(wert);
#endif
}


// Umwandlungsroutinen Digit sequence --> Integer:

// Normalized Digit sequence to Integer
// NDS_to_I(MSDptr,len)
// Digit Sequence MSDptr/len/.. in Integer umwandeln.
  extern cl_I NDS_to_I (const uintD* MSDptr, uintC len);

// Normalized Unsigned Digit Sequence to Integer
// NUDS_to_I(MSDptr,len)
// Normalized UDS MSDptr/len/.. in Integer >=0 umwandeln.
// Unterhalb von MSDptr mu 1 Digit Platz sein.
  extern cl_I NUDS_to_I (uintD* MSDptr, uintC len);

// Unsigned Digit Sequence to Integer
// UDS_to_I(MSDptr,len)
// UDS MSDptr/len/.. in Integer >=0 umwandeln.
// Unterhalb von MSDptr mu 1 Digit Platz sein.
  extern cl_I UDS_to_I (uintD* MSDptr, uintC len);

// Digit Sequence to Integer
// DS_to_I(MSDptr,len)
// DS MSDptr/len/.. in Integer umwandeln.
  extern cl_I DS_to_I (const uintD* MSDptr, uintC len);


// Umwandlungsroutinen Integer --> Digit sequence:

// Unterteilung eines Fixnums in Digits:
// intDsize=8 -> MSD=LSD3,LSD2,LSD1,LSD0, sollte FN_maxlength=4 sein.
// intDsize=16 -> MSD=LSD1,LSD0, sollte FN_maxlength=2 sein.
// intDsize=32 -> MSD=LSD0, sollte FN_maxlength=1 sein.

inline sintD FN_MSD (cl_uint word)
{
	// This assumes cl_value_shift + cl_value_len == cl_word_size.
	return (cl_sint)word >> (cl_value_shift + (FN_maxlength-1)*intDsize);
}

#if (FN_maxlength==1)
  #define FN_LSD0(word)  FN_MSD(word)
  extern uintD FN_LSD1 (cl_uint word); // never used, never defined
  extern uintD FN_LSD2 (cl_uint word); // never used, never defined
  extern uintD FN_LSD3 (cl_uint word); // never used, never defined
#endif
#if (FN_maxlength==2)
  inline uintD FN_LSD0 (cl_uint word)
  {
	return (uintD)(word >> cl_value_shift);
  }
  #define FN_LSD1(word)  FN_MSD(word)
  extern uintD FN_LSD2 (cl_uint word); // never used, never defined
  extern uintD FN_LSD3 (cl_uint word); // never used, never defined
#endif
#if (FN_maxlength==4)
  inline uintD FN_LSD0 (cl_uint word)
  {
	return (uintD)(word >> cl_value_shift);
  }
  inline uintD FN_LSD1 (cl_uint word)
  {
	return (uintD)(word >> (cl_value_shift + intDsize));
  }
  inline uintD FN_LSD2 (cl_uint word)
  {
	return (uintD)(word >> (cl_value_shift + 2*intDsize));
  }
  #define FN_LSD3(word)  FN_MSD(word)
#endif

// wird nur bei FN_maxlength >= 2 gebraucht, d.h. intDsize < cl_value_len
#define FN_MSD1_mask  (~((cl_uint)(bitc(intDsize-1)-1) << cl_value_shift))
// wird nur bei FN_maxlength >= 3 gebraucht, d.h. 2*intDsize < cl_value_len
#define FN_MSD2_mask  (~((cl_uint)(bitc(2*intDsize-1)-1) << cl_value_shift))
// wird nur bei FN_maxlength >= 4 gebraucht, d.h. 3*intDsize < cl_value_len
#define FN_MSD3_mask  (~((cl_uint)(bitc(3*intDsize-1)-1) << cl_value_shift))

// Store a Fixnum in destptr[-FN_maxlength,...,-1].
#define FN_to_NDS(destptr, word, MSDptr_zuweisung,len_zuweisung,LSDptr_zuweisung)  \
{ var uintC len_from_FN_to_NDS;						\
  if (word==0)								\
    { len_from_FN_to_NDS = 0; }						\
    else								\
    { var cl_uint testMSD;						\
      if ((FN_maxlength<=1) ||						\
          ((testMSD = (word) & FN_MSD1_mask) == 0) || (testMSD == FN_MSD1_mask) \
         )								\
        { len_from_FN_to_NDS = 1;					\
          (destptr)[-1] = FN_LSD0(word);				\
        }								\
      elif ((FN_maxlength<=2) ||					\
            ((testMSD = (word) & FN_MSD2_mask) == 0) || (testMSD == FN_MSD2_mask) \
           )								\
        { len_from_FN_to_NDS = 2;					\
          (destptr)[-1] = FN_LSD0(word);				\
          (destptr)[-2] = FN_LSD1(word);				\
        }								\
      elif ((FN_maxlength<=3) ||					\
            ((testMSD = (word) & FN_MSD3_mask) == 0) || (testMSD == FN_MSD3_mask) \
           )								\
        { len_from_FN_to_NDS = 3;					\
          (destptr)[-1] = FN_LSD0(word);				\
          (destptr)[-2] = FN_LSD1(word);				\
          (destptr)[-3] = FN_LSD2(word);				\
        }								\
      else /* (FN_maxlength<=4) */					\
        { len_from_FN_to_NDS = 4;					\
          (destptr)[-1] = FN_LSD0(word);				\
          (destptr)[-2] = FN_LSD1(word);				\
          (destptr)[-3] = FN_LSD2(word);				\
          (destptr)[-3] = FN_LSD3(word);				\
        }								\
    }									\
  unused (MSDptr_zuweisung (destptr) - (uintP)len_from_FN_to_NDS);	\
  len_zuweisung len_from_FN_to_NDS;					\
  unused (LSDptr_zuweisung (destptr));					\
}

// Bignum to Normalized Digit sequence, Kopieren unntig
// BN_to_NDS_nocopy(obj, MSDptr=,len=,LSDptr=);
// > obj: ein Bignum
// < MSDptr/len/LSDptr: Normalized Digit sequence
  #define BN_to_NDS_nocopy(obj, MSDptr_zuweisung,len_zuweisung,LSDptr_zuweisung)  \
    { var Bignum bn_from_BN_to_NDS_nocopy = TheBignum(obj);		\
      unused (MSDptr_zuweisung (const uintD*) &bn_from_BN_to_NDS_nocopy->data[0]);	\
      unused (LSDptr_zuweisung (const uintD*) &bn_from_BN_to_NDS_nocopy->data[(uintP)( len_zuweisung bn_from_BN_to_NDS_nocopy->length )]);		\
    }

// Bignum to Normalized Digit sequence
// BN_to_NDS(obj, MSDptr=,len=,LSDptr=);
// > obj: ein Bignum
// < MSDptr/len/LSDptr: Normalized Digit sequence, darf modifiziert werden.
// Dabei wird num_stack erniedrigt.
  #define BN_to_NDS(obj, MSDptr_zuweisung,len_zuweisung,LSDptr_zuweisung)  \
    { var const cl_I& obj_from_BN_to_NDS = (obj);			\
      var uintD* MSDptr_from_BN_to_NDS;					\
      var uintC len_from_BN_to_NDS;					\
      len_zuweisung len_from_BN_to_NDS = TheBignum(obj_from_BN_to_NDS)->length; \
      num_stack_need(len_from_BN_to_NDS, MSDptr_zuweisung MSDptr_from_BN_to_NDS = , LSDptr_zuweisung); \
      copy_loop_up(&TheBignum(obj_from_BN_to_NDS)->data[0],MSDptr_from_BN_to_NDS,len_from_BN_to_NDS); \
    }

// Bignum to Normalized Digit sequence
// BN_to_NDS_1(obj, MSDptr=,len=,LSDptr=);
// > obj: ein Bignum
// < MSDptr/len/LSDptr: Normalized Digit sequence, darf modifiziert werden.
// Unterhalb von MSDptr ist noch 1 Digit Platz.
// Dabei wird num_stack erniedrigt.
  #define BN_to_NDS_1(obj, MSDptr_zuweisung,len_zuweisung,LSDptr_zuweisung)  \
    { var const cl_I& obj_from_BN_to_NDS = (obj);			\
      var uintD* MSDptr_from_BN_to_NDS;					\
      var uintC len_from_BN_to_NDS;					\
      len_zuweisung len_from_BN_to_NDS = TheBignum(obj_from_BN_to_NDS)->length; \
      num_stack_need_1(len_from_BN_to_NDS, MSDptr_zuweisung MSDptr_from_BN_to_NDS = , LSDptr_zuweisung); \
      copy_loop_up(&TheBignum(obj_from_BN_to_NDS)->data[0],MSDptr_from_BN_to_NDS,len_from_BN_to_NDS); \
    }

// Integer to Normalized Digit sequence, Kopieren unntig.
// I_to_NDS_nocopy(obj, MSDptr=,len=,LSDptr=);
// > obj: ein Integer
// < MSDptr/len/LSDptr: Normalized Digit sequence
  #define I_to_NDS_nocopy(obj, MSDptr_zuweisung,len_zuweisung,LSDptr_zuweisung)  \
    var uintD CONCAT(FN_store_,__LINE__) [FN_maxlength];		\
    { var const cl_I& obj_from_I_to_NDS_nocopy = (obj);			\
      if (fixnump(obj_from_I_to_NDS_nocopy))				\
        { FN_to_NDS(& CONCAT(FN_store_,__LINE__) [FN_maxlength], obj_from_I_to_NDS_nocopy.word, MSDptr_zuweisung,len_zuweisung, LSDptr_zuweisung); } \
        else								\
        { BN_to_NDS_nocopy(obj_from_I_to_NDS_nocopy,MSDptr_zuweisung,len_zuweisung, LSDptr_zuweisung); } \
    }

// Integer to Normalized Digit sequence
// I_to_NDS(obj, MSDptr=,len=,LSDptr=);
// > obj: ein Integer
// < MSDptr/len/LSDptr: Normalized Digit sequence, darf modifiziert werden.
// Dabei wird num_stack erniedrigt.
  #define I_to_NDS(obj, MSDptr_zuweisung,len_zuweisung,LSDptr_zuweisung)  \
    var uintD CONCAT(FN_store_,__LINE__) [FN_maxlength];		\
    { var const cl_I& obj_from_I_to_NDS = (obj);			\
      if (fixnump(obj_from_I_to_NDS))					\
        { FN_to_NDS(& CONCAT(FN_store_,__LINE__) [FN_maxlength], obj_from_I_to_NDS.word, MSDptr_zuweisung,len_zuweisung, LSDptr_zuweisung); } \
        else                                                              \
        { BN_to_NDS(obj_from_I_to_NDS,MSDptr_zuweisung,len_zuweisung, LSDptr_zuweisung); } \
    }

// Integer to Normalized Digit sequence
// I_to_NDS_1(obj, MSDptr=,len=,LSDptr=);
// > obj: ein Integer
// < MSDptr/len/LSDptr: Normalized Digit sequence, darf modifiziert werden.
// Unterhalb von MSDptr ist noch 1 Digit Platz.
// Dabei wird num_stack erniedrigt.
  #define I_to_NDS_1(obj, MSDptr_zuweisung,len_zuweisung,LSDptr_zuweisung)  \
    var uintD CONCAT(FN_store_,__LINE__) [1+FN_maxlength];		\
    { var const cl_I& obj_from_I_to_NDS = (obj);			\
      if (fixnump(obj_from_I_to_NDS))					\
        { FN_to_NDS(& CONCAT(FN_store_,__LINE__) [1+FN_maxlength], obj_from_I_to_NDS.word, MSDptr_zuweisung,len_zuweisung, LSDptr_zuweisung); } \
        else                                                              \
        { BN_to_NDS_1(obj_from_I_to_NDS,MSDptr_zuweisung,len_zuweisung, LSDptr_zuweisung); } \
    }


// Division ganzer Zahlen

// Dividiert zwei Integers x,y >=0 und liefert Quotient und Rest
// der Division x/y. Bei y=0 Error.
// cl_divide(x,y)
// > x,y: Integers >=0
// < q,r: Quotient q, Rest r
  extern cl_I_div_t cl_divide (const cl_I& x, const cl_I& y);

// Fehler, wenn Quotient keine ganze Zahl ist
  nonreturning_function(extern, cl_error_exquo, (const cl_I& x, const cl_I& y));


// ggT und kgV von Integers

  // Teilfunktion fr die Durchfhrung des Euklid-Algorithmus auf
  // den fhrenden Ziffern a' und b':
  // partial_gcd(a',b',&erg); mit a'>b'
  // liefert in erg: x1,y1,x2,y2 mit den oben angegebenen Invarianten.
  typedef struct { uintD x1,y1,x2,y2; } partial_gcd_result;
  extern void partial_gcd (uintD z1, uintD z2, partial_gcd_result* erg);


// Wurzel aus ganzen Zahlen

// Stellt fest, ob ein Integer >=0 eine n-te Potenz ist.
// cl_rootp_aux(x,n,&w)
// > x: ein Integer >=0
// > n: ein Integer >0
// > Annahme: x > 1 und n < (integer-length x).
// < w: Integer (expt x (/ n)) falls x eine n-te Potenz
// < ergebnis: cl_true         ........................, cl_false sonst
  extern cl_boolean cl_rootp_aux (cl_I x, uintL n, cl_I* w);


// Hilfsfunktion zur Eingabe von Integers

// Wandelt eine Ziffernfolge in ein Integer >=0 um.
// digits_to_I(MSBptr,len,base)
// > base: Stellenwertsystem-Basis, >=2, <=36
// > MSBptr/len/..: Ziffernfolge, bestehend aus Punkten (werden berlesen)
//     und Ziffern/Buchstaben mit Wert < base.
// < ergebnis: der dargestellte Integer >=0
  extern cl_I digits_to_I (const char * MSBptr, uintL len, uintD base);


// A fixnum (cl_FN) is an immediate integer.

// typedef
class cl_FN : public cl_I {
public:
// Optimization of method pointer_p().
	cl_boolean pointer_p() const
		{ return cl_false; }
};

inline cl_boolean fixnump (const cl_FN& x)
	{ unused x; return cl_true; }
inline cl_boolean bignump (const cl_FN& x)
	{ unused x; return cl_false; }

inline cl_boolean minusp (const cl_FN& x)
{
	// This assumes cl_value_shift + cl_value_len == cl_word_size.
	return (cl_boolean)((cl_sint) x.word < 0);
}


// A bignum (cl_BN) is a heap-allocated integer.

// typedef
class cl_BN : public cl_I {
public:
// Optimization of method pointer_p().
	cl_boolean pointer_p() const
		{ return cl_true; }
};

inline cl_boolean fixnump (const cl_BN& x)
	{ unused x; return cl_false; }
inline cl_boolean bignump (const cl_BN& x)
	{ unused x; return cl_true; }

inline cl_boolean minusp (const cl_BN& x)
{
	return (cl_boolean)((sintD)TheBignum(x)->data[0] < 0);
}
inline cl_boolean zerop (const cl_BN& x)
	{ unused x; return cl_false; }

#endif /* _CL_I_H */
