// Input functions.

#ifndef _CL_INPUT_H
#define _CL_INPUT_H

#include "cl_number.h"
#include "cl_float.h"

// Wandelt eine Zeichenkette mit Integer-Syntax in ein Integer um.
// Punkte werden berlesen.
// read_integer(base,sign,string,index1,index2)
// > base: Lesebasis (>=2, <=36)
// > sign: Vorzeichen (/=0 falls negativ)
// > string: Simple-String (enthlt Ziffern mit Wert <base und evtl. Punkt)
// > index1: Index der ersten Ziffer
// > index2: Index nach der letzten Ziffer
//   (also index2-index1 Ziffern, incl. evtl. Dezimalpunkt am Schlu)
// < ergebnis: Integer
extern cl_I read_integer (unsigned int base,
            cl_signean sign, const char * string, uintL index1, uintL index2);

// Wandelt eine Zeichenkette mit Rational-Syntax in eine rationale Zahl um.
// read_rational(base,sign,string,index1,index3,index2)
// > base: Lesebasis (>=2, <=36)
// > sign: Vorzeichen (/=0 falls negativ)
// > string: Simple-String (enthlt Ziffern mit Wert <base und Bruchstrich)
// > index1: Index der ersten Ziffer
// > index3: Index von '/'
// > index2: Index nach der letzten Ziffer
//   (also index3-index1 Zhler-Ziffern, index2-index3-1 Nenner-Ziffern)
// < ergebnis: rationale Zahl
extern cl_RA read_rational (unsigned int base,
             cl_signean sign, const char * string, uintL index1, uintL index3, uintL index2);

struct cl_read_float_flags {
	cl_float_format_t default_float_format;
	cl_float_format_t default_lfloat_format;
};

// Wandelt eine Zeichenkette mit Float-Syntax in ein Float um.
// read_float(base,sign,string,index1,index4,index2,index3)
// > base: Lesebasis (=10)
// > sign: Vorzeichen (/=0 falls negativ)
// > string: Simple-String (enthlt Ziffern und evtl. Punkt und Exponentmarker)
// > index1: Index vom Mantissenanfang (excl. Vorzeichen)
// > index4: Index nach dem Mantissenende
// > index2: Index beim Ende der Characters
// > index3: Index nach dem Dezimalpunkt (=index4 falls keiner da)
//   (also Mantisse mit index4-index1 Characters: Ziffern und max. 1 '.')
//   (also index4-index3 Nachkommaziffern)
//   (also bei index4<index2: index4 = Index des Exponent-Markers,
//    index4+1 = Index des Exponenten-Vorzeichens oder der ersten
//    Exponenten-Ziffer)
// < ergebnis: Float
extern cl_F read_float (unsigned int base, const cl_read_float_flags& flags,
            cl_signean sign, const char * string, uintL index1, uintL index4, uintL index2, uintL index3);

// Specifies the possible results of a read operation.
enum cl_read_syntax_t {
	syntax_integer = bit(0),				// -> cl_I
	syntax_ratio = bit(1),					// -> cl_RA
	syntax_rational = syntax_integer | syntax_ratio,	// -> cl_RA
	syntax_sfloat = bit(2),					// -> cl_SF
	syntax_ffloat = bit(3),					// -> cl_FF
	syntax_dfloat = bit(4),					// -> cl_DF
	syntax_lfloat = bit(5),					// -> cl_LF
	syntax_float = syntax_sfloat | syntax_ffloat | syntax_dfloat | syntax_lfloat,
								// -> cl_F
	syntax_real = syntax_rational | syntax_float,		// -> cl_R
	syntax_complex = bit(6),				// -> cl_N
	syntax_number = syntax_real | syntax_complex,		// -> cl_N
	syntax_maybe_bad = bit(7),				// avoid errors
};

// Specifies the syntax to be applied to a read operation.
enum cl_read_lsyntax_t {
		// Standard algebraic notation.
	lsyntax_standard = 0,
		// Extended algebraic notation: x+yi
	lsyntax_algebraic = bit(0),
		// Common Lisp notation: #b, #o, #x, #r, #c
	lsyntax_commonlisp = bit(1),
		// All of them.
	lsyntax_all = lsyntax_algebraic | lsyntax_commonlisp,
};

struct cl_read_flags {
	cl_read_syntax_t syntax;
	cl_read_lsyntax_t lsyntax;
	unsigned int base;
	cl_read_float_flags float_flags;
};

extern cl_number read_number (const cl_read_flags& flags, const char * string, const char * string_limit, const char * * end_of_parse);

#endif /* _CL_INPUT_H */
