// Output functions.

#ifndef _CL_OUTPUT_H
#define _CL_OUTPUT_H

#include "cl_number.h"
#include "cl_io.h"

// Liefert einen String.
// Mit cl_malloc_hook() alloziert, mit cl_free_hook() freizugeben.
extern char * cl_string (const char * ptr, uintL len);

// Liefert zu einem Integer >=0  (write-to-string integer :base 10 :radix nil),
// also die Ziffernfolge als String.
// Mit cl_malloc_hook() alloziert, mit cl_free_hook() freizugeben.
extern char * cl_decimal_string (const cl_I& x);

// Gibt ein Integer aus.
// print_integer(stream,base,z);
// > z: Integer
// > base: Basis (>=2, <=36)
// > stream: Stream
extern void print_integer (cl_ostream stream, unsigned int base, const cl_I& z);

// Gibt eine rationale Zahl aus.
// print_rational(stream,base,z);
// > z: rationale Zahl
// > base: Basis (>=2, <=36)
// > stream: Stream
extern void print_rational (cl_ostream stream, unsigned int base, const cl_RA& z);

// Gibt ein Float aus.
// print_float(stream,z);
// > z: Float
// > stream: Stream
extern void print_float (cl_ostream stream, const cl_F& z);
// Flag whether to prefer type specific exponent markers over 'E'.
extern cl_boolean cl_print_float_readably;

// Gibt ein Float binr (sehr primitiv) aus.
// print_float_binary(stream,z);
// > z: Float
// > stream: Stream
extern void print_float_binary (cl_ostream stream, const cl_F& z);

// Gibt eine reelle Zahl aus.
// print_real(stream,z);
// > z: reelle Zahl
// > stream: Stream
extern void print_real (cl_ostream stream, const cl_R& z);
// Base in which rational numbers are to be printed.
extern unsigned int print_rational_base;
// Flag whether to print radix specifiers in Common Lisp syntax for
// rational numbers (#nR or #b or #x prefixes, trailing dot).
extern cl_boolean cl_print_rational_readably;

// Gibt eine komplexe Zahl aus.
// print_complex(stream,z);
// > z: komplexe Zahl
// > stream: Stream
extern void print_complex (cl_ostream stream, const cl_N& z);
// Flag whether to use the Common Lisp #C(realpart imagpart) syntax,
extern cl_boolean cl_print_complex_readably;

// Gibt eine Zahl aus.
// print_number(stream,z);
// > z: Zahl
// > stream: Stream
extern void print_number (cl_ostream stream, const cl_number& z);

#endif /* _CL_OUTPUT_H */
